//////////////////////////////////////////////////////////////////////
////                                                              ////
////  OR1200's Time Management Unit                               ////
////                                                              ////
////  This file is not part of the OpenRISC 1200 project          ////
////  Developed as a master thesis project at NTNU, 2014          ////
////                                                              ////
////  Description                                                 ////
////  Measures task execution time and arrival of interrupts	   ////
////                                                              ////
////                                                              ////
////                                                              ////
////  Author(s):                                                  ////
////      - Lars Ødegaard                                         ////
////      -Kyrre Gonsholt                                         ////
////                                                              ////
//////////////////////////////////////////////////////////////////////
`include "or1200_defines.v"

module or1200_tmu (
	//clk and rst
	clk, rst
	
	// SPR interface
	, spr_addr, spr_dat_i, spr_cs, spr_write, spr_dat_o
	
	//exception interface
	, intr, except_started
	
	//Interrupt interface
	, pic_ints, pic_ints_masked
	
	, ex_freeze

);

input				clk;
input				rst;
//
// SPR interface
//
input 	[31:0]		spr_addr;
input 	[31:0]		spr_dat_i;
input						spr_cs;
input						spr_write;
output 	[31:0]		spr_dat_o;
//
// Exception interface
//
input 					except_started;
output					intr;
//
//	Interrupt interface
//
input	[`OR1200_PIC_INTS-1:0] pic_ints;
output	[`OR1200_PIC_INTS-1:0] pic_ints_masked;

input						ex_freeze;


`ifdef OR1200_TMU_IMPLEMENTED

//
//	TMU status register
//
reg		[31:0]		sr;

//
//	TMU ctrl register
//
reg		[31:0]		ctrl;

//
//	TMU count register
//
reg		[63:0]		count;

//
//	TMU compare register
//
reg		[63:0]		compare;

//
//	Interrupt replenishment counter
//
reg		[63:0]		rep_count;
reg		[63:0]		rep_compare;

//
//	Interrupt flank register
//
wire		[`OR1200_PIC_INTS-1:0]	intr_edge;
reg		[`OR1200_PIC_INTS-1:0] 	intr_single;

//
//	Interrupt counting regisers
//
reg		[`OR1200_PIC_INTS-1:0]		pic_mask;
reg		[31:0]		intr_compare	[0:`OR1200_PIC_INTS-1];
reg		[31:0]		intr_count		[0:`OR1200_PIC_INTS-1];

//
//	Internal wires and regs
//
wire						sr_sel;
wire						comp_hi_sel;
wire						comp_lo_sel;
wire						count_hi_sel;
wire						count_lo_sel;
wire						repcomp_hi_sel;
wire						repcomp_lo_sel;
wire						repcnt_hi_sel;
wire						repcnt_lo_sel;

wire	[`OR1200_PIC_INTS-1:0] intr_count_sel;
wire	[`OR1200_PIC_INTS-1:0] intr_compare_sel;
reg		[31:0]				spr_dat_o;

wire						start;
wire						stop;
wire						restart;
wire						clear;
wire						ee;
wire						ce;
wire						fcd;
wire						ac;
wire						intr; /*synthesis keep*/
wire						replenish;

reg						running;
reg						count_invalid;
reg						suspended;

//
//	Decode register address
//
assign sr_sel 		= (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_STATUS)) ? 1'b1 : 1'b0;
assign ctrl_sel 	= (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_CTRL)) ? 1'b1 : 1'b0;
assign comp_hi_sel	= (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_COMPARE_HI)) ? 1'b1 : 1'b0;
assign comp_lo_sel	= (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_COMPARE_LO)) ? 1'b1 : 1'b0;
assign count_hi_sel = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_COUNT_HI)) ? 1'b1 : 1'b0;
assign count_lo_sel = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_COUNT_LO)) ? 1'b1 : 1'b0;
assign repcomp_hi_sel = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_REPCOMP_HI)) ? 1'b1 : 1'b0;
assign repcomp_lo_sel = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_REPCOMP_LO)) ? 1'b1 : 1'b0;
assign repcnt_hi_sel = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_REPCNT_HI)) ? 1'b1 : 1'b0;
assign repcnt_lo_sel = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_REPCNT_LO)) ? 1'b1 : 1'b0;


generate
genvar pic_nr;
	for (pic_nr = 0; pic_nr<`OR1200_PIC_INTS; pic_nr = pic_nr+1) begin : INTR_SEL
		assign intr_compare_sel[pic_nr] = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_INTRCOMP+pic_nr)) ? 1'b1 : 1'b0;
		assign intr_count_sel[pic_nr] = (spr_cs && (spr_addr[`OR1200_TMUOFS_BITS] == `OR1200_TMU_OFS_INTRCNT+pic_nr)) ? 1'b1 : 1'b0;
	end
endgenerate


//
//	Write to ctrl
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		ctrl <= 32'b0;
	end else if (ctrl_sel && spr_write) begin
		ctrl <= spr_dat_i;
	end else begin
		ctrl = 32'b0;
	end
end

//
//	Write to sr
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		//Set default operation
		sr[`OR1200_TMU_SR_R] <= 1'b0;
		sr[`OR1200_TMU_SR_EE] <= 1'b1;
		sr[`OR1200_TMU_SR_CE] <= 1'b0;
		sr[`OR1200_TMU_SR_FC] <= 1'b0;
		sr[`OR1200_TMU_SR_CI] <= 1'b1;
		sr[`OR1200_TMU_SR_CNTI] <= 1'b0;
		sr[`OR1200_TMU_SR_SUS] <= 1'b0;
		sr[`OR1200_TMU_SR_AC] <= 1'b1;
		
		sr[`OR1200_TMU_SR_UNUSED] <= 0;
	end else begin
		sr[`OR1200_TMU_SR_R] <= running;
		sr[`OR1200_TMU_SR_EE] <= ee;
		sr[`OR1200_TMU_SR_CE] <= ce;
		sr[`OR1200_TMU_SR_FC] <= fcd;
		sr[`OR1200_TMU_SR_CI] <= ci;
		sr[`OR1200_TMU_SR_CNTI] <= count_invalid;
		sr[`OR1200_TMU_SR_SUS] <= suspended;
		sr[`OR1200_TMU_SR_AC] <= ac;
		
		sr[`OR1200_TMU_SR_UNUSED] <= 0;
	end
end


//
//	Write to compare and count
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		compare <= 64'b0;
		count <= 64'b0;
		count_invalid <= 1'b0;
	//Count register
	end else if (running && !suspended) begin
		count <= count + 64'b1;
	end else if (count_hi_sel && spr_write && !running) begin
		count[`OR1200_TMU_HI_BITS] <= spr_dat_i;
		count_invalid <= 0;
	end else if (count_lo_sel && spr_write && !running) begin
		count[`OR1200_TMU_LO_BITS]	<= spr_dat_i;
		count_invalid <= 0;
	end else if (clear == 1'b1) begin
		count <= 64'b0;
		compare <= 64'b0;
		count_invalid <= 0;
	end else if (restart == 1'b1) begin
		count <= 64'b0;
		count_invalid <= 0;
	//Compare register
	end else if (comp_hi_sel && spr_write && !running) begin
		compare[`OR1200_TMU_HI_BITS] <= spr_dat_i;
		count_invalid <= 1'b1;
	end else if (comp_lo_sel && spr_write && !running) begin
		compare[`OR1200_TMU_LO_BITS] <= spr_dat_i;
		count_invalid <= 1'b1;
	end
end


//
//	Read TMU registers
//
always @(spr_addr or sr or compare or count) begin
		case (spr_addr[`OR1200_TMUOFS_BITS])
			`OR1200_TMU_OFS_STATUS: spr_dat_o = sr;
			`OR1200_TMU_OFS_CTRL: spr_dat_o = ctrl;
			`OR1200_TMU_OFS_COMPARE_HI: spr_dat_o = compare[`OR1200_TMU_HI_BITS];
			`OR1200_TMU_OFS_COMPARE_LO: spr_dat_o = compare[`OR1200_TMU_LO_BITS];
			`OR1200_TMU_OFS_COUNT_HI: spr_dat_o = count[`OR1200_TMU_HI_BITS];
			`OR1200_TMU_OFS_COUNT_LO: spr_dat_o = count[`OR1200_TMU_LO_BITS];
			`OR1200_TMU_OFS_REPCOMP_LO: spr_dat_o = rep_compare[`OR1200_TMU_LO_BITS];
			`OR1200_TMU_OFS_REPCOMP_HI: spr_dat_o = rep_compare[`OR1200_TMU_HI_BITS];
			`OR1200_TMU_OFS_REPCNT_LO: spr_dat_o = rep_count[`OR1200_TMU_HI_BITS];
			`OR1200_TMU_OFS_REPCNT_HI: spr_dat_o = rep_count[`OR1200_TMU_HI_BITS];
			`OR1200_TMU_OFS_PIC_MASK: begin
				spr_dat_o[31:`OR1200_PIC_INTS] = {32-`OR1200_PIC_INTS{1'b0}};
				spr_dat_o[`OR1200_PIC_INTS-1:0] = pic_mask;
			end
			`OR1200_TMU_OFS_INTRCNT: spr_dat_o = intr_count[0];
			`OR1200_TMU_OFS_INTRCNT+1: spr_dat_o = intr_count[1];
			`OR1200_TMU_OFS_INTRCNT+2: spr_dat_o = intr_count[2];
			`OR1200_TMU_OFS_INTRCNT+3: spr_dat_o = intr_count[3];
			`OR1200_TMU_OFS_INTRCNT+4: spr_dat_o = intr_count[4];
			`OR1200_TMU_OFS_INTRCNT+5: spr_dat_o = intr_count[5];
			`OR1200_TMU_OFS_INTRCNT+6: spr_dat_o = intr_count[6];
			`OR1200_TMU_OFS_INTRCNT+7: spr_dat_o = intr_count[7];
			`OR1200_TMU_OFS_INTRCNT+8: spr_dat_o = intr_count[8];
			`OR1200_TMU_OFS_INTRCNT+9: spr_dat_o = intr_count[9];
			`OR1200_TMU_OFS_INTRCNT+10: spr_dat_o = intr_count[10];
			`OR1200_TMU_OFS_INTRCNT+11: spr_dat_o = intr_count[11];
			`OR1200_TMU_OFS_INTRCNT+12: spr_dat_o = intr_count[12];
			`OR1200_TMU_OFS_INTRCNT+13: spr_dat_o = intr_count[13];
			`OR1200_TMU_OFS_INTRCNT+14: spr_dat_o = intr_count[14];
			`OR1200_TMU_OFS_INTRCNT+15: spr_dat_o = intr_count[15];
			`OR1200_TMU_OFS_INTRCNT+16: spr_dat_o = intr_count[16];
			`OR1200_TMU_OFS_INTRCNT+17: spr_dat_o = intr_count[17];
			`OR1200_TMU_OFS_INTRCNT+18: spr_dat_o = intr_count[18];
			//`OR1200_TMU_OFS_INTRCNT+19: spr_dat_o = intr_count[19];
			//`OR1200_TMU_OFS_INTRCNT+20: spr_dat_o = intr_count[20];
			//`OR1200_TMU_OFS_INTRCNT+21: spr_dat_o = intr_count[21];
			//`OR1200_TMU_OFS_INTRCNT+22: spr_dat_o = intr_count[22];
			//`OR1200_TMU_OFS_INTRCNT+23: spr_dat_o = intr_count[23];
			//`OR1200_TMU_OFS_INTRCNT+24: spr_dat_o = intr_count[24];
			//`OR1200_TMU_OFS_INTRCNT+25: spr_dat_o = intr_count[25];
			//`OR1200_TMU_OFS_INTRCNT+26: spr_dat_o = intr_count[26];
			//`OR1200_TMU_OFS_INTRCNT+27: spr_dat_o = intr_count[27];
			//`OR1200_TMU_OFS_INTRCNT+28: spr_dat_o = intr_count[28];
			//`OR1200_TMU_OFS_INTRCNT+29: spr_dat_o = intr_count[29];
			//`OR1200_TMU_OFS_INTRCNT+30: spr_dat_o = intr_count[30];
			//`OR1200_TMU_OFS_INTRCNT+31: spr_dat_o = intr_count[31];
			default: spr_dat_o = sr;
		endcase
end		


//
//	Decode of running
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		running <= 1'b0;
	end else if (intr == 1'b1) begin
		running <= 1'b0;
	end else if (running == 1'b0 && start == 1'b1 && compare > 0) begin
		running <= 1'b1;
	end else if (running == 1'b1 && stop == 1'b1) begin
		running <= 1'b0;
	end else if (running == 1'b1 && ce == 1'b0 && except_started == 1'b1) begin
		running <= 1'b0;
	end else if (running == 1'b1) begin
		running <= 1'b1;
	end else begin
		running <= 1'b0;
	end
end

//
//	Decode of suspended
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		suspended <= 1'b0;
	end else if (fcd == 1'b1 && ex_freeze == 1'b1) begin
		suspended <= 1'b1;
	end else if (ci == 1'b0 && 1'b1) begin
		suspended <= 1'b1;
	end else begin
		suspended <= 1'b0;
	end
end

//
//	Write to replenishment compare
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		rep_compare <= 64'b0;
	end else if (repcomp_hi_sel && spr_write) begin
		rep_compare[`OR1200_TMU_HI_BITS] <= spr_dat_i;
	end else if (repcomp_lo_sel && spr_write) begin
		rep_compare[`OR1200_TMU_LO_BITS] <= spr_dat_i;
	end
end


//
//	Write to interrupt compare registers
//
integer piccm;
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		for (piccm = 0; piccm < `OR1200_PIC_INTS; piccm = piccm + 1) begin
			intr_compare[piccm] <= 32'b0;
		end
	end else begin
		for (piccm = 0; piccm < `OR1200_PIC_INTS; piccm = piccm + 1) begin
			if (intr_compare_sel[piccm] && spr_write) begin
				intr_compare[piccm] <= spr_dat_i;
			end
		end 
	end
end
						
						
//
//	Increment interrupt counters
//
integer pics;
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		for (pics = 0; pics < `OR1200_PIC_INTS; pics = pics + 1) begin
			intr_count[pics] <= 32'b0;
		end
	end else if (replenish == 1'b1) begin
		for (pics = 0; pics < `OR1200_PIC_INTS; pics = pics + 1) begin
			intr_count[pics] <= 32'b0;
		end
	end else begin
		for (pics = 0; pics < `OR1200_PIC_INTS; pics = pics + 1) begin
			if (intr_edge[pics] == 1'b1 && sr[`OR1200_TMU_SR_CI]) begin
				intr_count[pics] <= intr_count[pics] + 1;
			end else begin
				intr_count[pics] <= intr_count[pics];
			end
		end
		pics = 0;
	end
end

//
//	Edge detector for interrupts
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		intr_single <= `OR1200_PIC_INTS'b0;
	end else begin
		intr_single <= pic_ints;
	end
end
assign intr_edge = pic_ints & (~intr_single);

//
//	Decode interrupt mask
//
integer mask;
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		pic_mask <= `OR1200_PIC_INTS'b0;
		mask <= 0;
	end else begin
		for (mask = 0; mask < `OR1200_PIC_INTS; mask = mask + 1) begin
			if ((intr_count[mask] >= intr_compare[mask]) && (intr_compare[mask] != 0) && sr[`OR1200_TMU_SR_CI]) begin
				pic_mask[mask] <= 1'b1;
			end else begin
				pic_mask[mask] <= 1'b0;
			end
		end
	end
end

//
//	Increment replenishment counter
//
always @(posedge clk or `OR1200_RST_EVENT rst) begin
	if (rst == `OR1200_RST_VALUE) begin
		rep_count <= 64'b0;
	end else if (replenish) begin
		rep_count <= 64'b0;
	end else if (rep_compare != 0) begin
		rep_count <= rep_count + 1;
	end else begin
		rep_count <= 64'b0;
	end
end


						
assign ee = ctrl[`OR1200_TMU_CTRL_EE] & ~ctrl[`OR1200_TMU_CTRL_ED] |
						sr[`OR1200_TMU_SR_EE] & ~ctrl[`OR1200_TMU_CTRL_ED];
						
assign ce = ctrl[`OR1200_TMU_CTRL_CEE] & ~ctrl[`OR1200_TMU_CTRL_CED] |
						sr[`OR1200_TMU_SR_CE] & ~ctrl[`OR1200_TMU_CTRL_CED];
						
assign fcd = ctrl[`OR1200_TMU_CTRL_FCE] & ~ctrl[`OR1200_TMU_CTRL_FCD] |
						sr[`OR1200_TMU_SR_FC] & ~ctrl[`OR1200_TMU_CTRL_FCD];
						
assign ci = ctrl[`OR1200_TMU_CTRL_CIE] & ~ctrl[`OR1200_TMU_CTRL_CID] |
						sr[`OR1200_TMU_SR_CI] & ~ctrl[`OR1200_TMU_CTRL_CID];
						
assign ac = ctrl[`OR1200_TMU_CTRL_ACE] & ~ctrl[`OR1200_TMU_CTRL_ACD] |
						sr[`OR1200_TMU_SR_AC] & ~ctrl[`OR1200_TMU_CTRL_ACD];
						
assign start = ctrl[`OR1200_TMU_CTRL_START];
assign stop = ctrl[`OR1200_TMU_CTRL_STOP];
assign restart = ctrl[`OR1200_TMU_CTRL_RESTART];
assign clear = ctrl[`OR1200_TMU_CTRL_CLEAR];

assign replenish = (rep_count == rep_compare) ? 1'b1 : 1'b0;
assign intr = ((count >= compare) && (sr[`OR1200_TMU_SR_EE] == 1'b1) && compare != 0 ) ? 1'b1 : 1'b0;
assign pic_ints_masked = pic_ints & ~pic_mask;


`else

//When TMU is not implemented drive outputs as if it was inactive

assign intr = 1'b0;
assign spr_dat_o = 32'b0;
assign pic_ints_masked = pic_ints;



`endif //OR1200_TMU_IMPLEMENTED

endmodule




