/* intgen.c -- IntGen external peripheral

   Copyright (C) 2008 Embecosm Limited

   Contributor Jeremy Bennett <jeremy.bennett@embecosm.com>

   This file is part of OpenRISC 1000 Architectural Simulator.

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 3 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program.  If not, see <http://www.gnu.org/licenses/>. */

/* This program is commented throughout in a fashion suitable for processing
   with Doxygen. */


/* This is functional simulation of any external peripheral. It's job is to
 * trap accesses in a specific range, so that the simulator can drive an
 * external device.
 *
 * A note on endianess. All external communication is done using HOST
 * endianess. A set of functions are provided to convert between host and
 * model endianess (htoml, htoms, mtohl, mtohs).
 */

/* Autoconf and/or portability configuration */
#include "config.h"

/* System includes */
#include <stdlib.h>
#include <stdio.h>


/* Package includes */
#include "arch.h"
#include "sim-config.h"
#include "abstract.h"
#include "toplevel-support.h"
#include "sim-cmd.h"
#include "sched.h"
#include "pic.h"

#define INTGEN0_TO 0x0
#define INTGEN1_TO 0x1000000
#define INTGEN0_CLR 0x4
#define INTGEN1_CLR 0x100004

#define MAX_SPRS_PER_GRP_BITS (11)
#define SPRGROUP_PIC (9<< MAX_SPRS_PER_GRP_BITS)
#define SPR_PICSR (SPRGROUP_PIC + 2)

/*! State associated with the intgen device. */
struct dev_intgen
{

  /* Info about a particular transaction */

  enum
  {				/* Direction of the access */
	INTGEN_READ,
    INTGEN_WRITE
  } trans_direction;

  enum
  {				/* Size of the access */
    INTGEN_BYTE,
    INTGEN_HW,
    INTGEN_WORD
  } trans_size;

  uint32_t value;		/* The value to read/write */

  /* Configuration */

  int enabled;			/* Device enabled */
  int word_enabled;		/* Full word R/W allowed */
  char *name;			/* Name of the device */
  oraddr_t baseaddr;		/* Base address of device */
  uint32_t size;		/* Address space size (bytes) */
  int version;

  /* Values */
  int timeout[2];
  int active[2];
};

#define RAND_MAX 65535
//#define RAND_MAX 268435455


static void intgen0(void *dat)
{
	struct dev_intgen *dev = (struct dev_intgen *) dat;
	if (dev->version == 1){
		report_interrupt(3);
	}
	else if (dev->version == 2)
	{
	report_interrupt(3);
	dev->active[0] = 1;
	dev->timeout[0]= rand()%RAND_MAX;
	SCHED_ADD(intgen0, dat, dev->timeout[0]);
	}

}

static void intgen1(void *dat)
{
	struct dev_intgen *dev = (struct dev_intgen *) dat;
	if (dev->version == 1)
	{
		report_interrupt(4);
	}
	else if (dev->version == 2)
	{
	report_interrupt(4);
	dev->active[1] = 1;
	dev->timeout[1]= rand()%RAND_MAX;
	SCHED_ADD(intgen1, dat, dev->timeout[1]);
	}
}



/* --------------------------------------------------------------------------*/
/*!Read a full word from an external device

   Since this is a full word, the result must be converted to host endianess.

   @note We are passed the device address, but we must convert it to a full
         address for external use, to allow the single upcall handler to
         decode multiple intgen devices.

   @param[in] addr  The device address to read from (host endian).
   @param[in] dat   The device data structure.

   @return  The full word read (host endian).                                */
/* --------------------------------------------------------------------------*/
static uint32_t
intgen_read_word (oraddr_t  addr,
		   void     *dat)
{

  struct dev_intgen *dev = (struct dev_intgen *) dat;

  int ret = 0;
  if (dev->enabled != 1)
  {
	  printf("IntGen:: %s \t \n Read to disabled intgen device \n", dev->name);
	  return 0;
  }
  switch (addr)
   {
   case INTGEN0_TO:
 	  ret = dev->timeout[0];
 	  break;
   case INTGEN1_TO:
   	  ret = dev->timeout[1];
   	  break;
   case INTGEN0_CLR:
	   ret = dev->active[0];
	   break;
   case INTGEN1_CLR:
	  ret = dev->active[1];
	  break;
   }
  return ret;
}	/* intgen_read_word() */


/* --------------------------------------------------------------------------*/
/*!Write a full word to an external device

   Since this is a half word, the value must be converted from host endianess.

   @note We are passed the device address, but we must convert it to a full
         address for external use, to allow the single upcall handler to
         decode multiple intgen devices.

   @param[in] addr  The device address to write to (host endian).
   @param[in] value The full word value to write (host endian).
   @param[in] dat   The device data structure.                               */
/* --------------------------------------------------------------------------*/
static void
intgen_write_word (oraddr_t  addr,
		    uint32_t  value,
		    void     *dat)
{
  struct dev_intgen *dev = (struct dev_intgen *) dat;
  if (dev->enabled != 1)
   {
 	  printf("IntGen:: %s \n \t Write to disabled intgen device \n ", dev->name );
 	  return;
   }
  if (addr < 0 || addr > (dev->size))
  {
	  printf("IntGen:: %s \n \t Read to memory outside of range \n \t Address: %lx \n", dev->name, addr);
	  return;
  }
  switch (addr)
   {
   case INTGEN0_TO:
 	  dev->timeout[0]= value;
 	  dev->active[0] = 1;
	  SCHED_ADD(intgen0, dev,value);
 	  break;
   case INTGEN0_CLR:
	  clear_interrupt(3);
	  dev->timeout[0] = 0;
	  dev->active[0] = 0;
	  break;
   case INTGEN1_TO:
	  dev->timeout[1]= value;
	  dev->active[1] = 1;
 	  SCHED_ADD(intgen1, dev,value);
 	  break;
   case INTGEN1_CLR:
	  clear_interrupt(4);
	  dev->timeout[1] = 0;
	  dev->active[1] = 0;
	  break;
   }
}	/* intgen_write_word() */

/* Status report can only advise of configuration. */

static void
intgen_status (void *dat)
{
  struct dev_intgen *dev = (struct dev_intgen *) dat;

  PRINTF ("\nIntGen device \"%s\" at 0x%" PRIxADDR ":\n", dev->name, dev->baseaddr);
  PRINTF ("  Size 0x%" PRIx32 "\n", dev->size);

  if (dev->word_enabled)
    {
      PRINTF ("  Full word R/W enabled\n");
    }
  PRINTF("   Version %d \n ", dev->version);

  PRINTF ("\n");

}				/* intgen_status() */


/* Reset is a null operation */

static void
intgen_reset (void *dat)
{
	struct dev_intgen *dev = (struct dev_intgen *) dat;
	dev->timeout[0]= 0;
	dev->timeout[1]= 0;

	SCHED_ADD(intgen_status, dat,1);
}				/* intgen_reset() */

/* Functions to set configuration */

static void
intgen_enabled (union param_val val, void *dat)
{
  ((struct dev_intgen *) dat)->enabled = val.int_val;

}				/* intgen_enabled() */

static void
intgen_word_enabled (union param_val val, void *dat)
{
  ((struct dev_intgen *) dat)->word_enabled = val.int_val;

}				/* intgen_word_enabled() */

static void
intgen_baseaddr (union param_val val, void *dat)
{
  ((struct dev_intgen *) dat)->baseaddr = val.addr_val;

}				/* intgen_baseaddr() */


static void
intgen_size (union param_val val, void *dat)
{
  ((struct dev_intgen *) dat)->size = val.int_val;

}				/* intgen_size() */

static void
intgen_version(union param_val val, void *dat)
{
  ((struct dev_intgen *) dat)->version = val.int_val;

}				/* intgen_version() */

/* Start of new intgen section */

static void *
intgen_sec_start ()
{
  struct dev_intgen *new =
    (struct dev_intgen *) malloc (sizeof (struct dev_intgen));

  if (0 == new)
    {
      fprintf (stderr, "IntGen peripheral: Run out of memory\n");
      exit (-1);
    }

  /* Default names */

  new->enabled = 1;
  new->word_enabled = 1;
  new->name = "Interrupt Generator";
  new->baseaddr = 0xa0000000;
  new->size = 0x01000008;
  new->version = 1;

  return new;

}				/* intgen_sec_start() */


/* End of new intgen section */

static void
intgen_sec_end (void *dat)
{
  struct dev_intgen *intgen = (struct dev_intgen *) dat;
  struct mem_ops ops;

  /* Give up if not enabled, or if size is zero, or if no access size is
     enabled. */

  if (!intgen->enabled)
    {
      free (dat);
      return;
    }

  if (0 == intgen->size)
    {
      fprintf (stderr, "IntGen peripheral \"%s\" has size 0: ignoring",
	       intgen->name);
      free (dat);
      return;
    }

  if (!intgen->word_enabled)
    {
      fprintf (stderr, "IntGen peripheral \"%s\" has no access: ignoring",
	       intgen->name);
      free (dat);
      return;
    }

  /* Zero all the ops, then set the ones we care about. Read/write delays will
   * come from the peripheral if desired.
   */

  memset (&ops, 0, sizeof (struct mem_ops));

  if (intgen->word_enabled)
    {
      ops.readfunc32 = intgen_read_word;
      ops.writefunc32 = intgen_write_word;
      ops.read_dat32 = dat;
      ops.write_dat32 = dat;
    }

  /* Register everything */
  reg_mem_area (intgen->baseaddr, intgen->size, 0, &ops);

  reg_sim_reset (intgen_reset, dat);
  reg_sim_stat (intgen_status, dat);

}				/* intgen_sec_end() */


/* Register a intgen section. */

void
reg_intgen_sec (void)
{
  struct config_section *sec = reg_config_sec ("intgen",
					       intgen_sec_start,
					       intgen_sec_end);

  reg_config_param (sec, "enabled",      PARAMT_INT, intgen_enabled);
  reg_config_param (sec, "word_enabled", PARAMT_INT, intgen_word_enabled);
  reg_config_param (sec, "baseaddr",     PARAMT_ADDR, intgen_baseaddr);
  reg_config_param (sec, "size",         PARAMT_INT, intgen_size);
  reg_config_param (sec, "version",      PARAMT_INT, intgen_version);

}				/* reg_intgen_sec */
